/* This file is part of the Dreamcast function library.
 * Please see libdream.c for further details.
 *
 * (c)2000 Dan Potter
 */

#include "dream.h"

/* Checks the attached cable type (to the A/V port). Returns
   one of the following:
     0 == VGA
     1 == (nothing)
     2 == RGB
     3 == Composite
   
   This is a direct port of Marcus' assembly function of the
   same name. */
int dc_vid_check_cable() {
	unsigned long * porta = (unsigned long *)0xff80002c;
	
	*porta = (*porta & 0xfff0ffff) | 0x000a0000;
	
	// Read port8 and port9
	return (*((unsigned short*)(porta + 1)) >> 8) & 3;
}

/* The base address of the video chip */
unsigned long * videobase = (unsigned long *)0xa05f8000;

/* Base address of the video ram */
unsigned long * vram_l = (unsigned long *)0xa5000000;
unsigned short * vram_s = (unsigned short *)0xa5000000;

/* Pixel configuration (set by init_video) */
int vram_config = 0;

/* Size of the video memory (set by init_video) */
int vram_size = 0;


/* Set up the video registers to the desired video mode (only
   640x480 supported right now). This function does not currently
   initialize all registers, but assumes that the boot ROM has set
   up reasonable defaults for syncs, etc.

   This is a fairly direct port of Marcus' assembly function of the 
   same name.
   
   cable_type (0=VGA, 2=RGB, 3=Composite)
   pixel_mode (0=RGB555, 1=RGB565, 3=RGB888)
   
   */
void dc_vid_init(int cable_type, int pixel_mode) {
	unsigned long * cvbsbase = (unsigned long *)0xa0702c00;
	int disp_lines, size_modulo, il_flag;

	// Looks up bytes per pixel as a shift value
	unsigned char bppshifttab[] = { 1, 1, 0, 2 };
	unsigned long bppshift = bppshifttab[pixel_mode];

	// Save the video memory size
	vram_size = (640*480*4) << bppshift;
	vram_config = pixel_mode;
	
	// Set border color
	dc_vid_border_color(0, 0, 0);
	
	// Set pixel clock and color mode
	pixel_mode = (pixel_mode << 2) | 1;	// Color mode and DE
	disp_lines = (240 / 2) << 1;
	if (cable_type != 3) {
		// Double # of display lines for VGA and S-Video
		disp_lines <<= 1;		// Double line count
		pixel_mode |= 0x800000;		// Set clock double
	}
	*(videobase + 0x11) = pixel_mode;
	
	// Set video base address (right at the top)
	*(videobase + 0x14) = 0;
	
	// Video base address for short fields should be offset by one line
	*(videobase + 0x15) = 640 << bppshift;
	
	// Set screen size and modulo, and interlace flag
	il_flag = 0x100;	// Set Video Output Enable
	size_modulo = 0;
	if (cable_type != 0) {	// non-VGA => interlace
		// Add one line to offset => display every other line
		size_modulo = (640/4) << bppshift;
		il_flag |= 0x50;	// Enable interlace
	}
	*(videobase + 0x17) = ((size_modulo+1) << 20) | (disp_lines << 10) 
		| (((640/4) << bppshift) - 1);
	*(videobase + 0x34) = il_flag;
	
	// Set vertical pos and border
	*(videobase + 0x37) = (24 << 16) | (24 + disp_lines);
	*(videobase + 0x3c) = (24 << 16) | 24;
	
	// Horizontal pos
	*(videobase + 0x3b) = 0xa4;
	
	// Select RGB/CVBS
	if (cable_type == 3)
		*cvbsbase = 3 << 8;
	else
		*cvbsbase = 0;
}


/* Sets the border color of the Dreamcast display */
void dc_vid_border_color(int r, int g, int b) {
	*(videobase + 0x10) = (r << 16) | (g << 8) | (b << 0);
}

/* Clears the screen */
void dc_vid_clear(int r, int g, int b) {
	switch (vram_config) {
		case PM_RGB555: {
			unsigned short pixel = ((r >> 3) << 10)
				| ((g >> 3) << 5)
				| ((b >> 3) << 0);
			int i;
			for (i=0; i<vram_size; i++)
				vram_s[i] = pixel;
		} break;
		case PM_RGB565: {
			unsigned short pixel = ((r >> 3) << 11)
				| ((g >> 2) << 5)
				| ((b >> 3) << 0);
			int i;
			for (i=0; i<vram_size; i++)
				vram_s[i] = pixel;
		} break;
		case PM_RGB888: {
			unsigned long pixel = (r << 16) | (g << 8) | (b << 0);
			int i;
			for (i=0; i<vram_size; i++)
				vram_l[i] = pixel;
		} break;
	}
}


/* Waits for a vertical refresh to start. This is the period between
   when the scan beam reaches the bottom of the picture, and when it
   starts again at the top.
   
   Thanks to HeroZero for this info. */
void dc_vid_waitvbl() {
	volatile unsigned long *vbl = videobase + 0x010c/4;
	while (!(*vbl & 0x01ff))
		;
	while (*vbl & 0x01ff)
		;
}

